""" Tests of Markdown renderer """
# pylint:disable=missing-function-docstring

from publ import markdown


def test_render_title():
    assert markdown.render_title("This *is* a test") == "This <em>is</em> a test"
    assert markdown.render_title("This *is* a test", markup=False) == "This is a test"
    assert markdown.render_title("This is ~~not~~ a test", markup=False) == "This is a test"
    assert markdown.render_title("This is <s>not</s> a test", markup=False) == "This is a test"
    assert markdown.render_title("This & That", markup=False) == "This & That"
    assert markdown.render_title("This & That", markup=True) == "This &amp; That"

    assert markdown.render_title('The "sun" is a liberal myth',
                                 markup=False, smartquotes=False) == 'The "sun" is a liberal myth'
    assert markdown.render_title('The "sun" is a liberal myth',
                                 markup=False, smartquotes=True) == 'The “sun” is a liberal myth'
    assert markdown.render_title('The "sun" is a liberal myth',
                                 markup=True, smartquotes=False) == \
        'The &quot;sun&quot; is a liberal myth'
    assert markdown.render_title('The "sun" is a liberal myth',
                                 markup=True, smartquotes=True) == \
        'The &ldquo;sun&rdquo; is a liberal myth'



""" tests of publ.utils module """
# pylint:disable=missing-function-docstring

import flask
import pytest

from publ import utils


def test_callable_proxy():
    """ Tests of the CallableProxy class """

    def wrapped_func(retval="default"):
        return retval

    proxy = utils.CallableProxy(wrapped_func)

    assert proxy() == "default"
    assert proxy(1) == 1
    assert proxy(True) is True
    assert proxy(False) is False

    app = flask.Flask(__name__)
    with app.test_request_context("/"):
        assert proxy.upper() == "DEFAULT"
        assert bool(proxy)
        assert len(proxy) == 7
        assert str(proxy) == "default"
        assert list(proxy) == ['d', 'e', 'f', 'a', 'u', 'l', 't']
        assert proxy[3] == 'a'

        assert proxy == "default"
        assert 't' in proxy
        assert proxy + "foo" == "defaultfoo"


def test_callable_proxy_caching():
    """ Test the caching-ness of CallableProxy's default """

    app = flask.Flask(__name__)
    app.secret_key = 'bogus'

    stash = {'call_count': 0}

    def wrapped_func():
        stash['call_count'] += 1
        return flask.request.path

    proxy = utils.CallableProxy(wrapped_func)

    with app.test_request_context("/foo/bar"):
        # within same path the call should only happen once
        assert str(proxy) == "/foo/bar"
        assert stash['call_count'] == 1
        assert str(proxy) == "/foo/bar"
        assert stash['call_count'] == 1

    with app.test_request_context("/foo/bar"):
        # cached value should remain cached
        assert str(proxy) == "/foo/bar"
        assert stash['call_count'] == 1

    with app.test_request_context("/foo/baz"):
        # different path should call it again
        assert str(proxy) == "/foo/baz"
        assert stash['call_count'] == 2

    with app.test_request_context("/foo/baz"):
        # different user should call it again
        flask.session['me'] = 'test:cachebuster'
        assert str(proxy) == "/foo/baz"
        assert stash['call_count'] == 3


def test_truecallable_proxy():
    """ Tests of TrueCallableProxy """

    stash = {'call_count': 0}

    def wrapped_func():
        stash['call_count'] += 1
        return False

    proxy = utils.TrueCallableProxy(wrapped_func)

    app = flask.Flask(__name__)
    with app.test_request_context('/foo'):
        assert stash['call_count'] == 0
        assert proxy
        assert stash['call_count'] == 0


def test_callablevalue():
    """ Tests of CallableValue """
    proxy = utils.CallableValue("flonk")
    falsy = utils.CallableValue(False)

    app = flask.Flask(__name__)
    with app.test_request_context("/bar"):
        assert proxy == "flonk"
        assert len(proxy) == 5
        assert proxy + "foo" == "flonkfoo"
        assert proxy
        assert not falsy


def test_tagset_membership():
    """ Membership tests for TagSet """
    from publ.utils import TagSet

    items = ('a', 'S', 'd', 'F')
    others = ('Q', 'w', 'E', 'r')
    tags = TagSet(items)

    assert bool(tags)
    assert not bool(TagSet())
    assert not bool(TagSet([]))
    assert not bool(TagSet({}))

    assert len(tags) == 4

    assert hash(TagSet(('a', 's', 'd'))) == hash(TagSet(('s', 'd', 'a')))
    assert hash(TagSet(('a', 's', 'd'))) == hash(TagSet(('A', 'S', 'D')))
    assert hash(TagSet(items)) != hash(TagSet(others))

    for item in items:
        assert item in tags
        assert item.lower() in tags
        assert item.upper() in tags
        assert item.casefold() in tags

    for item in tags:
        assert item.casefold() in {t.casefold() for t in items}

    for item in others:
        assert item not in tags
        assert item.lower() not in tags
        assert item.upper() not in tags
        assert item.casefold() not in tags


def test_tagset_operators():
    """ Test operators  on TagSet """
    from publ.utils import TagSet

    assert TagSet(('a', 's', 'D', 'F')) == TagSet(('A', 'S', 'd', 'f'))
    assert TagSet(('a', 's', 'D', 'F')) == {'A', 'S', 'd', 'f'}
    assert TagSet(('a', 's', 'd', 'f')) != TagSet(('a', 's', 'd'))
    assert TagSet(('a', 's', 'd', 'f')) != {'a', 's', 'd'}

    assert TagSet(('a', 'S', 'd')) | TagSet(('A', 's', 'D', 'F')) == TagSet(('a', 's', 'd', 'f'))
    assert TagSet(('a', 'S', 'd')) | {'A', 's', 'D', 'F'} == TagSet(('a', 's', 'd', 'f'))

    assert TagSet(('a', 'S', 'd')) & TagSet(('a', 'D')) == TagSet(('A', 'd'))
    assert TagSet(('a', 'S', 'd')) & {'a', 'D'} == TagSet(('A', 'd'))

    assert TagSet(('a', 's', 'd', 'f')) ^ TagSet(
        ('A', 'F', 'G', 'h')) == TagSet(('s', 'd', 'g', 'h'))
    assert TagSet(('a', 's', 'd', 'f')) ^ {'A', 'F', 'G', 'h'} == TagSet(('s', 'd', 'g', 'h'))

    assert TagSet(('a', 's', 'd', 'f')) - {'A', 'D', 'g', 'G'} == {'s', 'f'}

    assert TagSet(('1', '2', '3')) < TagSet(('1', '2', '3', '4'))
    assert TagSet(('1', '2', '3')) < {'1', '2', '3', '4'}

    assert TagSet(('1', '2', '3')) <= TagSet(('1', '2', '3', '4'))
    assert TagSet(('1', '2', '3')) <= {'1', '2', '3', '4'}

    # pylint:disable=unneeded-not
    assert not TagSet(('a', 's', 'd')) < TagSet(('a', 's', 'd'))
    assert not TagSet(('a', 's', 'd')) < TagSet(('q', 'w', 'e'))
    assert not TagSet(('a', 's', 'd')) <= TagSet(('q', 'w', 'e'))


def test_parse_tuple_string():
    """ test of parse_tuple_string() """
    assert utils.parse_tuple_string((1, 2, 3)) == (1, 2, 3)
    assert utils.parse_tuple_string([1, 2, 3]) == (1, 2, 3)
    assert utils.parse_tuple_string("1,2,3") == (1, 2, 3)
    assert utils.parse_tuple_string("a,,", bool) == (True, False, False)
    with pytest.raises(Exception):
        utils.parse_tuple_string("a,b,c", float)
    assert utils.parse_tuple_string(None) is None


def test_make_tag():
    """ test of utils.make_tag """

    assert utils.make_tag('a', {'href': 'foo'}, False) == '<a href="foo">'
    assert utils.make_tag('a',
                          [('href', 'foo'), ('href', 'bar')],
                          True) == '<a href="foo" href="bar" />'
    with pytest.raises(Exception):
        utils.make_tag('a', (('href', 'foo')))

    app = flask.Flask(__name__)
    with app.test_request_context():
        proxy = utils.CallableValue("<hello>")
        assert utils.make_tag('a', {'href': proxy}) == '<a href="&lt;hello&gt;">'

    escaped = flask.Markup("&amp;")
    assert utils.make_tag('a', {'href': escaped}) == '<a href="&amp;">'


def test_listlike():
    """ test functions involving the ListLike inference class """
    assert utils.is_list((1, 2, 3))
    assert utils.is_list([])
    assert utils.is_list([1, 2, 3])
    assert utils.is_list({1, 2, 3})
    assert utils.is_list(frozenset((1, 2, 3)))
    assert utils.is_list(utils.TagSet(('1', '2', 'A', 'a')))
    assert not utils.is_list("")
    assert not utils.is_list("foo")
    assert not utils.is_list(123)
    assert not utils.is_list(True)
    assert not utils.is_list(None)

    assert utils.as_list([1, 2, 3]) == [1, 2, 3]
    assert utils.as_list((1, 2, 3)) == (1, 2, 3)
    assert utils.as_list("foo") == ("foo",)
    assert utils.as_list(True) == (True,)
    assert utils.as_list(None) == ()


def test_parse_date():
    """ tests for the date parser """
    import arrow
    from publ import config

    def make_date(year=None, month=None, day=None):
        return arrow.Arrow(year=year or 1,
                           month=month or 1,
                           day=day or 1,
                           tzinfo=config.timezone)

    with pytest.raises(ValueError):
        utils.parse_date("not a date")
    with pytest.raises(ValueError):
        utils.parse_date("12345678")
    with pytest.raises(ValueError):
        utils.parse_date("")

    assert utils.parse_date('1978-06-14') == (make_date(1978, 6, 14), 'day', utils.DAY_FORMAT)
    assert utils.parse_date('19780614') == (make_date(1978, 6, 14), 'day', utils.DAY_FORMAT)

    assert utils.parse_date('1983-07') == (make_date(1983, 7), 'month', utils.MONTH_FORMAT)
    assert utils.parse_date('198307') == (make_date(1983, 7), 'month', utils.MONTH_FORMAT)

    assert utils.parse_date("1979") == (make_date(1979), 'year', utils.YEAR_FORMAT)

    assert utils.parse_date('19810505_w') == (make_date(1981, 5, 4), 'week', utils.WEEK_FORMAT)


def test_find_file():
    """ tests for the file finder """
    assert utils.find_file("anything", []) is None

    assert utils.find_file("anything", ["tests/templates"]) is None
    assert utils.find_file("auth", ["tests/templates"]) is None

    assert utils.find_file("index.html",
                           ("tests/templates")
                           ) == "tests/templates/index.html"
    assert utils.find_file("index.html",
                           ("tests/templates", "tests/templates/auth")
                           ) == "tests/templates/index.html"
    assert utils.find_file("index.html",
                           ("tests/templates/auth", "tests/templates")
                           ) == "tests/templates/auth/index.html"


def test_static_url():
    """ tests for the static URL builder """
    app = flask.Flask("tests", static_folder="asdf")
    with app.test_request_context("https://foo.bar/poiupoiupoiu"):
        assert utils.static_url("thing", absolute=False) == "/asdf/thing"
        assert utils.static_url("thong", absolute=True) == "https://foo.bar/asdf/thong"


def test_remap_link_target():
    """ test the target remapper """
    app = flask.Flask("tests", static_folder="feedme")
    with app.test_request_context("https://feed.me/blah/merry"):
        assert utils.remap_link_target("@fred") == "/feedme/fred"
        assert utils.remap_link_target("@fred", absolute=True) == "https://feed.me/feedme/fred"
        assert utils.remap_link_target("daphne") == "daphne"
        assert utils.remap_link_target("daphne", absolute=True) == "https://feed.me/blah/daphne"


def test_get_category():
    import os.path
    assert utils.get_category(os.path.join("foo", "bar", "entry.html")) == "foo/bar"


def test_remap_args():
    """ test the argument parser remapper """
    args = {'foo': 'bar', 'quuz': 'quux'}
    assert utils.remap_args(args, {}) == args
    assert utils.remap_args(args, {"foo": "poiu"}) == args
    assert utils.remap_args(args, {"foo": "quuz"})['foo'] == 'quux'


def test_prefix_normalize():
    """ test the prefix remapping thing """
    args = {'here': '1',
            'prefix_here':
            '2', 'prefix_there': '3',
            'anywhere': '4',
            'prefix': 'prefix_'}
    assert utils.prefix_normalize(args) == {
        'here': '2',
        'there': '3',
        'anywhere': '4',
        'prefix': 'prefix_'
    }



""" tests of publ.html_entry module """
# pylint:disable=missing-function-docstring


def test_process_passthrough():
    from publ.html_entry import process

    passthrough = '''<!DOCTYPE html>
<html><head>
<link rel="alternate" href="//example.com/" />
<title>This entry should pass through unscathed</title></head>
<body><h1>This entry should not be modified at all.</h1>

<p>This entry is one of those ones which shouldn't have anything that gets
modified in it, and shouldn't require an
<a href="https://flask.palletsprojects.com/en/1.1.x/appcontext/">application
context</a> to function.</p>

<img src="//example.com/some-image.png" width="500">
<img data-qwer="poiu" width="yes" height="no">

<br/><br/>

<!-- commentary -->

</body></html>'''
    assert process(passthrough, {}, ()) == passthrough

    assert process('<img data-publ-rewritten src="do_not_rewrite.jpg" width=400 height=400>',
                   {}, ()) == '<img src="do_not_rewrite.jpg" width="400" height="400">'


def test_process_attr_rewrites():
    import flask
    from publ.html_entry import process

    app = flask.Flask(__name__, static_folder="bleh")
    with app.test_request_context("https://foo.bar/baz"):
        assert process('<a href="@something">foo</a>', {}, ()) == \
            '<a href="/bleh/something">foo</a>'
        assert process('<a href="@something">bar</a>', {'absolute': True}, ()) == \
            '<a href="https://foo.bar/bleh/something">bar</a>'

        assert process('<div $data-something="@something" />', {}, ()) == \
            '<div data-something="/bleh/something" />'
        assert process('<div $data-something="@something" />', {'absolute': True}, ()) == \
            '<div data-something="https://foo.bar/bleh/something" />'


def test_image_args():
    import flask
    from publ.html_entry import process

    app = flask.Flask(__name__, static_folder="bleh")
    with app.test_request_context("https://foo.bar/baz"):
        assert process('<img src="//example.com/image.png{500}">', {}, ()) == \
            '<img src="//example.com/image.png" width="500">'


def test_process_strip_html():
    from publ.html_entry import process

    assert process('<a href="foo">bar</a>', {'markup': False}, ()) == "bar"


def test_strip_html():
    from publ.html_entry import strip_html

    assert strip_html("foobar") == "foobar"

    doc = '<a href="zxcv" class="mew">blah<sup>boo</sup></a><br/>'

    assert strip_html(doc) == "blahboo"

    assert strip_html(doc, ('sup')) == "blah<sup>boo</sup>"

    assert strip_html(doc, ('a'), ('href')) == '<a href="zxcv">blahboo</a>'

    assert strip_html(doc, remove_elements=('sup')) == 'blah'

    assert strip_html(doc, ('br')) == 'blahboo<br/>'

    assert strip_html("this &amp; that") == "this & that"



# markdown.py
""" markdown formatting functionality """

import logging
import re
import typing
import urllib.parse

import flask
import misaka
import pygments
import pygments.formatters
import pygments.lexers
import slugify

from . import config, html_entry, image, links, utils

LOGGER = logging.getLogger(__name__)

TocEntry = typing.Tuple[int, str]
TocBuffer = typing.List[TocEntry]

# Allow these tags in TOC entries
TOC_ALLOWED_TAGS = ('sup', 'sub',
                    'em', 'strong',
                    'b', 'i',
                    'code',
                    'del', 'add', 'mark')

# Remove these tags from plaintext-style conversions
PLAINTEXT_REMOVE_ELEMENTS = ('del', 's')


class HtmlRenderer(misaka.HtmlRenderer):
    """ Customized renderer for enhancing Markdown formatting

    Constructor arguments:

    config -- The configuration for the Markdown tags
    search_path -- Directories to look in for resolving relatively-linked files
    """

    def __init__(self, args: typing.Dict,
                 search_path: typing.Tuple[str],
                 entry_id: typing.Optional[int],
                 footnote_buffer: typing.Optional[typing.List[str]],
                 toc_buffer: typing.Optional[TocBuffer]):
        # pylint:disable=no-member,too-many-arguments
        super().__init__(0, args.get('xhtml') and misaka.HTML_USE_XHTML or 0)

        self._config = args
        self._search_path = search_path
        self._entry_id = entry_id

        self._footnote_ofs = len(footnote_buffer) if footnote_buffer is not None else 0
        self._footnote_buffer = footnote_buffer

        self._toc_buffer = toc_buffer

    @staticmethod
    def footnotes(_):
        """ Actual footnote rendering is handled by the caller """
        return None

    def _footnote_num(self, num):
        return num + self._footnote_ofs

    def _footnote_id(self, num, anchor):
        return '{anchor}_e{eid}_fn{num}'.format(
            anchor=anchor,
            eid=self._entry_id,
            num=self._footnote_num(num))

    def _footnote_url(self, num, anchor):
        return urllib.parse.urljoin(self._config.get('footnotes_link', ''),
                                    '#' + self._footnote_id(num, anchor))

    def footnote_ref(self, num):
        """ Render a link to this footnote """
        if self._config.get('_suppress_footnotes'):
            return '\u200b'  # zero-width space to prevent Misaka fallback

        return '{sup}{link}{content}</a></sup>'.format(
            sup=utils.make_tag('sup', {
                'id': self._footnote_id(num, "r"),
                'class': self._config.get('footnotes_class', False)
            }),
            link=utils.make_tag('a', {
                'href': self._footnote_url(num, "d"),
                'rel': 'footnote'
            }),
            content=self._footnote_num(num))

    def footnote_def(self, content, num):
        """ Render the footnote body, deferring it if so configured """
        LOGGER.debug("footnote_def %d: %s", num, content)

        # Insert the return anchor before the end of the first content block
        before, partition, after = content.partition('</p>')
        text = '{li}{before}&nbsp;{link}{icon}</a>{partition}{after}</li>'.format(
            li=utils.make_tag('li', {
                'id': self._footnote_id(num, "d")
            }),
            before=before,
            link=utils.make_tag('a', {
                'href': self._footnote_url(num, "r"),
                'rev': 'footnote'
            }),
            icon=self._config.get('footnotes_return', '↩'),
            partition=partition,
            after=after,
        )

        self._footnote_buffer.append(text)

    def _header_id(self, content, level, num):
        """ Return a reasonable anchor ID for a heading """
        return '{eid}_h{level}_{num}_{slug}'.format(
            eid=self._entry_id,
            level=level,
            num=num,
            slug=slugify.slugify(content, max_length=32))

    def header(self, content, level):
        """ Make a header with anchor """

        htag = 'h{level}'.format(level=level)
        hid = self._header_id(html_entry.strip_html(content,
                                                    remove_elements=PLAINTEXT_REMOVE_ELEMENTS),
                              level, len(self._toc_buffer) + 1)

        atag = utils.make_tag('a', {
            'href': urllib.parse.urljoin(self._config.get('toc_link', ''),
                                         '#' + hid),
            'class': self._config.get('heading_link_class', False),
            **self._config.get('heading_link_config', {})
        })

        if self._toc_buffer is not None:
            LOGGER.debug("append toc: %d %s", level, content)
            self._toc_buffer.append(
                (level,
                 '{atag}{content}</a>'.format(
                     atag=atag,
                     content=html_entry.strip_html(content, allowed_tags=TOC_ALLOWED_TAGS))))

        content = self._config.get('heading_template', '{link}</a>{text}').format(
            link=atag,
            text=content)

        return '{htag_open}{content}</{htag}>'.format(
            htag_open=utils.make_tag(htag, {'id': hid}),
            content=content,
            htag=htag)

    def image(self, raw_url, title='', alt=''):
        """ Adapt a standard Markdown image to a generated rendition set.

        Container arguments used (in addition to the rendition tags):

        div_class -- The CSS class name to use on any wrapper div
        div_style -- Additional CSS styles to apply to the wrapper div
        count -- The maximum number of images to show at once
        more_text -- If there are more than `count` images, add this text indicating
            that there are more images to be seen. This string gets two template
            arguments, `{count}` which is the total number of images in the set,
            and `{remain}` which is the number of images omitted from the set.
        more_link -- If `more_text` is shown, this will format the text as a link to this location.
        more_class -- If `more_text` is shown, wraps it in a `<div>` with this class.
        """
        # pylint: disable=too-many-locals

        text = ''

        image_specs = raw_url
        if title:
            image_specs += ' "{}"'.format(title)

        alt, container_args = image.parse_alt_text(alt)

        container_args = utils.prefix_normalize({**self._config, **container_args})

        spec_list = image.get_spec_list(image_specs, container_args)

        remain = 0
        for (spec, show) in spec_list:
            text += self._render_image(spec, show,
                                       container_args,
                                       alt)
            if not show:
                remain += 1

        if remain and 'more_text' in container_args:
            more_text = container_args['more_text'].format(
                count=len(spec_list),
                remain=remain)
            if 'more_link' in container_args:
                more_text = '{a}{text}</a>'.format(
                    text=more_text,
                    a=utils.make_tag('a', {'href': container_args['more_link']}))
            if 'more_class' in container_args:
                more_text = '{div}{text}</div>'.format(
                    text=more_text,
                    div=utils.make_tag('div', {'class': container_args['more_class']}))
            text += flask.Markup(more_text)

        if text and (container_args.get('div_class') or
                     container_args.get('div_style')):
            text = '{tag}{text}</div>'.format(
                tag=utils.make_tag('div',
                                   {'class': container_args.get('div_class') or False,
                                    'style': container_args.get('div_style') or False}),
                text=text)

        # if text is ''/falsy then misaka interprets this as a failed parse...
        return text or ' '

    def blockcode(self, text, lang):
        """ Pass a code fence through pygments """
        if lang and self._config.get('highlight_syntax', 'True'):
            try:
                lexer = pygments.lexers.get_lexer_by_name(lang, stripall=True)
            except pygments.lexers.ClassNotFound:
                lexer = None

            if lexer:
                formatter = pygments.formatters.HtmlFormatter()  # pylint: disable=no-member
                return pygments.highlight(text, lexer, formatter)

        return '\n<div class="highlight"><pre>{}</pre></div>\n'.format(
            flask.escape(text.strip()))

    def link(self, content, link, title=''):
        """ Emit a link, potentially remapped based on our embed or static rules """

        link = links.resolve(link, self._search_path,
                             self._config.get('absolute'))

        return '{}{}</a>'.format(
            utils.make_tag('a', {
                'href': link,
                'title': title if title else False
            }),
            content)

    @staticmethod
    def paragraph(content):
        """ emit a paragraph, stripping out any leading or following empty paragraphs """

        # if the content contains a top-level div then don't wrap it in a <p>
        # tag
        if content.startswith('<div') and content.endswith('</div>'):
            return '\n' + content + '\n'

        text = '<p>' + content + '</p>'
        text = re.sub(r'<p>\s*</p>', r'', text)
        return text or ' '

    def _render_image(self, spec, show, container_args, alt_text=None):
        """ Render an image specification into an <img> tag """

        try:
            path, image_args, title = image.parse_image_spec(spec)
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception("Got error on spec %s: %s", spec, err)
            return flask.Markup('<span class="error">Couldn\'t parse image spec: ' +
                                '<code>{}</code> {}</span>'.format(flask.escape(spec),
                                                                   flask.escape(str(err))))

        composite_args = {**container_args, **image_args}

        try:
            img = image.get_image(path, self._search_path)
            return img.get_img_tag(title,
                                   alt_text,
                                   **composite_args,
                                   _show_thumbnail=show,
                                   _mark_rewritten=True)
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception("Got error on image %s: %s", path, err)
            return flask.Markup('<span class="error">Error loading image {}: {}</span>'.format(
                flask.escape(spec), flask.escape(str(err))))


class ItemCounter(misaka.BaseRenderer):
    """ Just counts the number of things in an entry without any further processing """

    def __init__(self, toc: int = 0, footnote: int = 0):
        super().__init__()
        self.toc = toc
        self.footnote = footnote

    def __str__(self):
        return 'ItemCounter(footnote={footnote},toc={toc})'.format(footnote=self.footnote,
                                                                   toc=self.toc)

    def header(self, content, level):
        """ count this header """
        # pylint:disable=unused-argument
        self.toc += 1

    def footnote_def(self, content, num):
        """ count this footnote """
        # pylint:disable=unused-argument
        self.footnote += 1


def to_html(text, args, search_path,
            entry_id: typing.Optional[int] = None,
            toc_buffer: typing.Optional[TocBuffer] = None,
            footnote_buffer: typing.Optional[typing.List[str]] = None,
            postprocess: bool = True):
    # pylint:disable=too-many-arguments
    """ Convert Markdown text to HTML.

    toc_buffer -- a list of (level,text) for all headings in the entry

    footnote_buffer -- a list that will contain <li>s with the footnote items, if
    there are any footnotes to be found.

    postprocess -- whether to postprocess the buffers for smartypants/HTML/etc.
    """

    footnotes: typing.List[str] = footnote_buffer if footnote_buffer is not None else []
    tocs: TocBuffer = toc_buffer if toc_buffer is not None else []

    # first process as Markdown
    renderer = HtmlRenderer(args,
                            search_path,
                            toc_buffer=tocs,
                            footnote_buffer=footnotes,
                            entry_id=entry_id)
    processor = misaka.Markdown(renderer,
                                args.get('markdown_extensions', config.markdown_extensions))
    text = processor(text)

    if postprocess:
        # convert smartquotes, if so configured.
        # We prefer setting 'smartquotes' but we fall back to the negation of
        # 'no_smartquotes' for backwards compatibility with a not-well-considered
        # API.
        if 'no_smartquotes' in args:
            LOGGER.warning("no_smartquotes is deprecated and will be removed in a future version")
        smartquotes = args.get('smartquotes', not args.get('no_smartquotes', False))
        if smartquotes:
            text = misaka.smartypants(text)
            footnotes[:] = (misaka.smartypants(text) for text in footnotes)
            tocs[:] = ((level, misaka.smartypants(text)) for level, text in tocs)

        # now filter through html_entry to rewrite local src/href links
        text = html_entry.process(text, args, search_path)
        footnotes[:] = (html_entry.process(text, args, search_path)
                        for text in footnotes)

    return flask.Markup(text)


def get_counters(text, args):
    """ Count the number of stateful items in Markdown text. """
    counter = ItemCounter()
    processor = misaka.Markdown(counter,
                                args.get('markdown_extensions', config.markdown_extensions))
    processor(text)
    return counter


class TitleRenderer(HtmlRenderer):
    """ A renderer that is suitable for rendering out page titles and nothing else """

    def __init__(self):
        super().__init__({}, [], entry_id=0, toc_buffer=[], footnote_buffer=[])

    @staticmethod
    def paragraph(content):
        """ Passthrough """
        return content

    @staticmethod
    def list(content, is_ordered, is_block):
        """ Passthrough """
        # pylint: disable=unused-argument
        return content

    @staticmethod
    def listitem(content, is_ordered, is_block):
        """ Just add the * back on """
        # pylint: disable=unused-argument
        if not is_ordered:
            return '* ' + content
        raise ValueError("Not sure how we got here")

    @staticmethod
    def header(content, level):
        """ Passthrough """
        # pylint: disable=unused-argument
        return content


def render_title(text, markup=True, smartquotes=True, markdown_extensions=None):
    """ Convert a Markdown title to HTML """

    # If the title starts with something that looks like a list, save it for
    # later
    pfx, text = re.match(r'([0-9. ]*)(.*)', text).group(1, 2)

    text = pfx + misaka.Markdown(TitleRenderer(),
                                 extensions=markdown_extensions
                                 or config.markdown_extensions)(text)

    if smartquotes:
        text = misaka.smartypants(text)

    if not markup:
        text = html_entry.strip_html(text, remove_elements=PLAINTEXT_REMOVE_ELEMENTS)

    return flask.Markup(text)


def toc_to_html(toc: TocBuffer, max_level: int = None) -> str:
    """ Convert a TocBuffer to an appropriate <ol> """

    if not toc:
        return ''

    cur_level = 0
    out = ''

    # preprocess: find the lowest heading level
    min_level = min([level for level, _ in toc])
    toc = [(level - min_level + 1, text) for level, text in toc]

    for level, text in toc:
        if max_level is None or level <= max_level:
            if level > cur_level:
                # open sublists
                out += '<ol><li>' * (level - cur_level)
            else:
                # close sublists and the prior entry from this level
                out += '</li></ol>' * (cur_level - level) + '</li><li>'
            out += text
            cur_level = level

    out += '</li></ol>' * cur_level
    return out



""" Smoke test site runner """
# pylint:disable=invalid-name

import logging
import os

import authl.flask
import flask

import publ
import publ.image

APP_PATH = os.path.dirname(os.path.abspath(__file__))

logging.basicConfig(level=logging.DEBUG)

config = {
    'database_config': {
        'provider': 'sqlite',
        'filename': os.path.join(APP_PATH, '..', 'index.db')
    },
    'content_folder': os.path.join(APP_PATH, 'content'),
    'template_folder': os.path.join(APP_PATH, 'templates'),
    'static_folder': os.path.join(APP_PATH, 'static'),
    'cache': {
        'CACHE_TYPE': os.environ['TEST_CACHING'],
        'CACHE_DEFAULT_TIMEOUT': 600,
        'CACHE_THRESHOLD': 20
    } if os.environ.get('TEST_CACHING') else {
        'CACHE_NO_NULL_WARNING': True
    },
    'auth': {
        'TEST_ENABLED': True,

        'INDIEAUTH_CLIENT_ID': authl.flask.client_id,
        'INDIELOGIN_CLIENT_ID': authl.flask.client_id,

        'FEDIVERSE_NAME': 'Publ test suite',

        'TWITTER_CLIENT_KEY': os.environ.get('TWITTER_CLIENT_KEY'),
        'TWITTER_CLIENT_SECRET': os.environ.get('TWITTER_CLIENT_SECRET'),

        'EMAIL_SENDMAIL': print,
        'EMAIL_FROM': 'nobody@example.com',
        'EMAIL_SUBJECT': 'Log in to authl test',
        'EMAIL_CHECK_MESSAGE': 'Use the link printed to the test console',
    },
    'user_list': 'users.cfg',
}

app = publ.Publ(__name__, config)
app.secret_key = "We are insecure"


@app.route('/favicon.<ext>')
def favicon(ext):
    """ render a favicon """
    logo = publ.image.get_image('images/rawr.jpg', 'tests/content')
    img, _ = logo.get_rendition(format=ext, width=128, height=128, resize='fill')
    return flask.redirect(img)
